/*\
 * $Id: ppmtobmp.c,v 1.9 1992/11/24 19:39:33 dws Exp dws $
 * 
 * ppmtobmp.c - Converts from a PPM file to a Microsoft Windows or OS/2
 * .BMP file.

    Big hack attack by Kelvin to make Home Planet day/night files.

 * 
 * The current implementation is probably not complete, but it works for
 * me.	I welcome feedback.
 * 
 * Copyright (C) 1992 by David W. Sanderson.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  This software is provided "as is"
 * without express or implied warranty.
 * 
 * $Log: ppmtobmp.c,v $
 *
 * Hacked by John Walker to write 24 bit Windows BMP files if given an
 * input pixmap with more than 256 colours.  I don't know whether OS/2
 * supports an equivalent format or not.  I also changed the default
 * output file type from -os2 to -windows.  11 April 1993.
 *
 * Revision 1.9  1992/11/24  19:39:33  dws
 * Added copyright.
 *
 * Revision 1.8  1992/11/17  02:16:52  dws
 * Moved length functions to bmp.h.
 *
 * Revision 1.7  1992/11/11  23:18:16  dws
 * Modified to adjust the bits per pixel to 1, 4, or 8.
 *
 * Revision 1.6  1992/11/11  22:43:39  dws
 * Commented out a superfluous message.
 *
 * Revision 1.5  1992/11/11  05:58:06  dws
 * First version that works.
 *
 * Revision 1.4  1992/11/11  03:40:32  dws
 * Moved calculation of bits per pixel to BMPEncode.
 *
 * Revision 1.3  1992/11/11  03:02:34  dws
 * Added BMPEncode function.
 *
 * Revision 1.2  1992/11/08  01:44:35  dws
 * Added option processing and reading of PPM file.
 *
 * Revision 1.1  1992/11/08  00:46:07  dws
 * Initial revision
\*/

#include        "bmp.h"
#include        "ppm.h"
#include        "ppmcmap.h"
#include        "bitio.h"

#define MAXCOLORS 256

static pixval maxval;		    /* Maximum pixel value in input bitmap */

/*
 * Utilities
 */

static char     er_write[] = "stdout: write error";

static void
PutByte(fp, v)
	FILE	       *fp;
	char		v;
{
	if (putc(v, fp) == EOF)
	{
		pm_error(er_write);
	}
}

static void
PutShort(fp, v)
	FILE	       *fp;
	short		v;
{
	if (pm_writelittleshort(fp, v) == -1)
	{
		pm_error(er_write);
	}
}

static void
PutLong(fp, v)
	FILE	       *fp;
	long		v;
{
	if (pm_writelittlelong(fp, v) == -1)
	{
		pm_error(er_write);
	}
}

/*
 * BMP writing
 */

/*
 * returns the number of bytes written, or -1 on error.
 */
static int
BMPwritefileheader(fp, class, bitcount, x, y)
	FILE	       *fp;
	int		class;
	unsigned long	bitcount;
	unsigned long	x;
	unsigned long	y;
{
        PutByte(fp, 'B');
        PutByte(fp, 'M');

	/* cbSize */
	PutLong(fp, BMPlenfile(class, bitcount, x, y));

	/* xHotSpot */
	PutShort(fp, 0);

	/* yHotSpot */
	PutShort(fp, 0);

	/* offBits */
	PutLong(fp, BMPoffbits(class, bitcount));

	return 14;
}

/*
 * returns the number of bytes written, or -1 on error.
 */
static int
BMPwriteinfoheader(fp, class, bitcount, x, y)
	FILE	       *fp;
	int		class;
	unsigned long	bitcount;
	unsigned long	x;
	unsigned long	y;
{
	long	cbFix;

	/* cbFix */
	switch (class)
	{
	case C_WIN:
		cbFix = 40;
		PutLong(fp, cbFix);

		/* cx */
		PutLong(fp, x);
		/* cy */
		PutLong(fp, y);
		/* cPlanes */
		PutShort(fp, 1);
		/* cBitCount */
		PutShort(fp, bitcount);

		/*
                 * We've written 16 bytes so far, need to write 24 more
		 * for the required total of 40.
		 */

		PutLong(fp, 0);
		PutLong(fp, 0);
		PutLong(fp, 0);
		PutLong(fp, 0);
		PutLong(fp, 0);
		PutLong(fp, 0);


		break;
	case C_OS2:
		cbFix = 12;
		PutLong(fp, cbFix);

		/* cx */
		PutShort(fp, x);
		/* cy */
		PutShort(fp, y);
		/* cPlanes */
		PutShort(fp, 1);
		/* cBitCount */
		PutShort(fp, bitcount);

		break;
	default:
                pm_error(er_internal, "BMPwriteinfoheader");
	}

	return cbFix;
}

/*
 * returns the number of bytes written, or -1 on error.
 */
static int
BMPwritergb(fp,class,R,G,B)
	FILE	       *fp;
	int		class;
	pixval		R;
	pixval		G;
	pixval		B;
{
	switch (class)
	{
	case C_WIN:
		PutByte(fp, B);
		PutByte(fp, G);
		PutByte(fp, R);
		PutByte(fp, 0);
		return 4;
	case C_OS2:
		PutByte(fp, B);
		PutByte(fp, G);
		PutByte(fp, R);
		return 3;
	default:
                pm_error(er_internal, "BMPwritergb");
	}
	return -1;
}

/*
 * returns the number of bytes written, or -1 on error.
 */
static int
BMPwritergbtable(fp,class,bpp,colors,R,G,B)
	FILE	       *fp;
	int		class;
	int		bpp;
	pixval	       *R;
	pixval	       *G;
	pixval	       *B;
{
	int		nbyte = 0;
	int		i;
	long		ncolors;

	for (i = 0; i < colors; i++)
	{
		nbyte += BMPwritergb(fp,class,R[i],G[i],B[i]);
	}

	ncolors = (1 << bpp);

	for (; i < ncolors; i++)
	{
		nbyte += BMPwritergb(fp,class,0,0,0);
	}

	return nbyte;
}

/*
 * returns the number of bytes written, or -1 on error.
 */
static int
BMPwriterow(fp, row, cx, bpp, cht)
	FILE	       *fp;
	pixel	       *row;
	unsigned long	cx;
	unsigned short	bpp;
	colorhash_table cht;
{
	BITSTREAM	b;
	unsigned	nbyte = 0;
	int		rc;
	unsigned	x;

        if ((b = pm_bitinit(fp, "w")) == (BITSTREAM) 0)
	{
		return -1;
	}

	for (x = 0; x < cx; x++, row++)
	{
	    if (bpp == 24) {
		pixval pr, pg, pb;

		if (maxval != 255) {
		    pr = (pixval) (((int) PPM_GETR(*row)) * 255 / ((int) maxval));
		    pg = (pixval) (((int) PPM_GETG(*row)) * 255 / ((int) maxval));
		    pb = (pixval) (((int) PPM_GETB(*row)) * 255 / ((int) maxval));
		} else {
		    pr = PPM_GETR(*row);
		    pg = PPM_GETG(*row);
		    pb = PPM_GETB(*row);
		}
		if ((rc = pm_bitwrite(b, 8, pb)) == -1) {
		    return -1;
		}
		if ((rc = pm_bitwrite(b, 8, pg)) == -1) {
		    return -1;
		}
		if ((rc = pm_bitwrite(b, 8, pr)) == -1) {
		    return -1;
		}
		nbyte += 3;
	    } else {
		pixval pmx = 1 + (ppm_lookupcolor(cht, row) | 0x80);
		if ((rc = pm_bitwrite(b, bpp, pmx)) == -1)
		{
			return -1;
		}
		nbyte += rc;
	    }
	}

	if ((rc = pm_bitfini(b)) == -1)
	{
		return -1;
	}
	nbyte += rc;

	/*
	 * Make sure we write a multiple of 4 bytes.
	 */
	while (nbyte % 4)
	{
		PutByte(fp, 0);
		nbyte++;
	}

	return nbyte;
}

/*
 * returns the number of bytes written, or -1 on error.
 */
static int
BMPwritebits(fp, cx, cy, cBitCount, pixels, cht)
	FILE	       *fp;
	unsigned long	cx;
	unsigned long	cy;
	unsigned short	cBitCount;
	pixel	      **pixels;
	colorhash_table cht;
{
	int		nbyte = 0;
	long		y;

	if(cBitCount > 24)
	{
                pm_error("cannot handle cBitCount: %d"
			 ,cBitCount);
	}

	/*
	 * The picture is stored bottom line first, top line last
	 */

	for (y = cy - 1; y >= 0; y--)
	{
		int rc;
		rc = BMPwriterow(fp, pixels[y], cx, cBitCount, cht);

		if(rc == -1)
		{
                        pm_error("couldn't write row %d"
				 ,y);
		}
		if(rc%4)
		{
                        pm_error("row had bad number of bytes: %d"
				 ,rc);
		}
		nbyte += rc;
	}

	return nbyte;
}

/*
 * Return the number of bits per pixel required to represent the
 * given number of colors.
 */

static int
colorstobpp(colors)
	int		colors;
{
	int		bpp;

	if (colors < 1)
	{
                pm_error("can't have less than one color");
	}

	if ((bpp = pm_maxvaltobits(colors - 1)) > 8)
	{
                pm_error("can't happen");
	}

	return bpp;
}

/*
 * Write a BMP file of the given class.
 * 
 * Note that we must have 'colors' in order to know exactly how many
 * colors are in the R, G, B, arrays.  Entries beyond those in the
 * arrays are undefined.  If make24 is nonzero, we create a 24 bit
 * per pixel BMP file, ignoring colours, cht, and the R, G, and B
 * arrays.
 *
 */
static void
BMPEncode(fp, class, make24, x, y, pixels, colors, cht, R, G, B)
	FILE	       *fp;
	int		class;
	int		make24;
	int		x;
	int		y;
	pixel	      **pixels;
	int		colors; /* number of valid entries in R,G,B */
	colorhash_table cht;
	pixval	       *R;
	pixval	       *G;
	pixval	       *B;
{
	int		bpp;	/* bits per pixel */
	unsigned long	nbyte = 0;

	bpp = make24 ? 24 : colorstobpp(colors);

	/*
	 * I have found empirically at least one BMP-displaying program
         * that can't deal with (for instance) using 3 bits per pixel.
	 * I have seen no programs that can deal with using 3 bits per
	 * pixel.  I have seen programs which can deal with 1, 4, and
	 * 8 bits per pixel.
	 *
	 * Based on this, I adjust actual the number of bits per pixel
	 * as follows.	If anyone knows better, PLEASE tell me!
	 */
	switch(bpp)
	{
	case 2:
	case 3:
		bpp = 4;
		break;
	case 5:
	case 6:
	case 7:
		bpp = 8;
		break;
	}

        pm_message("Using %d bits per pixel", bpp);

	nbyte += BMPwritefileheader(fp, class, bpp, x, y);
	nbyte += BMPwriteinfoheader(fp, class, bpp, x, y);
	if (!make24) {
	    nbyte += BMPwritergbtable(fp, class, bpp, 256, R, G, B);
	}

	if(nbyte !=	( BMPlenfileheader(class)
			+ BMPleninfoheader(class)
			+ BMPlenrgbtable(class, bpp)))
	{
                pm_error(er_internal, "BMPEncode");
	}

	nbyte += BMPwritebits(fp, x, y, bpp, pixels, cht);
	if(nbyte != BMPlenfile(class, bpp, x, y))
	{
                pm_error(er_internal, "BMPEncode");
	}
}

int
main(argc, argv)
	int		argc;
	char	      **argv;
{
	FILE	       *ifp = stdin;
        char           *usage = "[-windows / -os2] [-rgb] [ppmfile]";
	int		class = C_WIN;	 /* Make default Windows, not OS/2 */
	int		make24 = 0;	 /* Create 24 bit BMP file ? */

	int		argn;
	int		rows;
	int		cols;
	int		colors;
	int		i;
	colorhist_vector chv;
	pixval		Red[MAXCOLORS];
	pixval		Green[MAXCOLORS];
	pixval		Blue[MAXCOLORS];

	pixel** pixels;
	colorhash_table cht;

	ppm_init(&argc, argv);

	argn = 1;

        while (argn < argc && argv[argn][0] == '-' && argv[argn][1] != '\0')
	{
                if (pm_keymatch(argv[argn], "-windows", 2))
			class = C_WIN;
                else if (pm_keymatch(argv[argn], "-os2", 2))
			class = C_OS2;
                else if (pm_keymatch(argv[argn], "-rgb", 2))
			make24 = 1;
		else
			pm_usage(usage);
		++argn;
	}

	if (argn < argc)
	{
		ifp = pm_openr(argv[argn]);
		++argn;
	}

	if (argn != argc)
	{
		pm_usage(usage);
	}

	pixels = ppm_readppm(ifp, &cols, &rows, &maxval);

	pm_close(ifp);

	if (!make24) {

	    /* Figure out the colormap. */

            pm_message("computing colormap...");
	    chv = ppm_computecolorhist(pixels, cols, rows, 126, &colors);
	    if (chv == (colorhist_vector) 0) {
                pm_error("More than 126 colours.  Cannot make WHUMP BMP");
	    } else {
                pm_message("%d colors found", colors);
	    }
	}

	/*
	 * Now turn the ppm colormap into the appropriate BMP
	 * colormap.
	 */
	if (maxval > 255)
	{
                pm_message("maxval is not 255 - automatically rescaling colors");
	}

	if (!make24) {
	    memset(Red, 0, sizeof Red);
	    memset(Green, 0, sizeof Green);
	    memset(Blue, 0, sizeof Blue);
	    for (i = 0; i < colors; ++i)
	    {
		    if (maxval == 255)
		    {
/*#define Subdue(x) ((pixval) ((((int) (x)) * 2) / 3))*/
#define Subdue(x) ((pixval) ((((int) (x)) * 1) / 2))
/*#define Subdue(x) ((x))*/
			    Red[i + 129] = PPM_GETR(chv[i].color);
			    Green[i + 129] = PPM_GETG(chv[i].color);
			    Blue[i + 129] = PPM_GETB(chv[i].color);
			    Red[i + 1] = Subdue(PPM_GETR(chv[i].color));
			    Green[i + 1] = Subdue(PPM_GETG(chv[i].color));
			    Blue[i + 1] = Subdue(PPM_GETB(chv[i].color));
		    }
		    else
		    {
			pixval ar, ag, ab;
			    ar = (pixval) (((int) PPM_GETR(chv[i].color)) * 255 / ((int) maxval));
			    ag = (pixval) (((int) PPM_GETG(chv[i].color)) * 255 / ((int) maxval));
			    ab = (pixval) (((int) PPM_GETB(chv[i].color)) * 255 / ((int) maxval));
			    Red[i + 128] = ar;
			    Green[i + 128] = ag;
			    Blue[i + 128] = ab;
			    Red[i] = Subdue(ar);
			    Green[i] = Subdue(ag);
			    Blue[i] = Subdue(ab);
		    }
	    }

	    /* And make a hash table for fast lookup. */
	    cht = ppm_colorhisttocolorhash(chv, colors);
	    ppm_freecolorhist(chv);
	}

        /* All set, let's do it. */
	BMPEncode(stdout, class, make24,
		cols, rows, pixels, 256, cht,
		Red, Green, Blue);

	pm_close(stdout);

	return 0;
}
